# Copyright 2004 James Bunton <james@delx.cjb.net>
# Licensed for distribution under the GPL version 2, check COPYING for details

import utils
import legacy
import jabw
import debug
import config




def makeSession(pytrans, jabberID):
	""" Tries to create a session object for the corresponding JabberID. Retrieves information
	from XDB to create the session. If it fails, then the user is most likely not registered with
	the transport """
	debug.log("session: makeSession() \"%s\"" % (jabberID))
	result = pytrans.registermanager.getRegInfo(jabberID)
	if(result):
		username, password, nickname = result
		return Session(pytrans, jabberID, username, password, nickname)
	else:
		return None



class Session(jabw.JabberConnection):
	""" A class to represent each registered user's session with the legacy network. Exists as long as there
	is a Jabber resource for the user available """
	
	def __init__(self, pytrans, jabberID, username, password, nickname):
		""" Initialises the session object and connects to the legacy network """
		jabw.JabberConnection.__init__(self, pytrans, jabberID)
		debug.log("Session: Creating new session \"%s\"" % (jabberID))
		
		self.pytrans = pytrans
		self.jabberID = jabberID # the JabberID of the Session's user
		self.username = username # the legacy network ID of the Session's user
		self.password = password
		self.nickname = nickname
		
		self.resourceList = []
		
		self.legacycon = legacy.LegacyConnection(self.username, self.password, self)
		
		# Remove when stable FIXME
		self.sendMessage(to=self.jabberID, fro=config.jid, body="This is an experimental gateway, PyMSN-t. If you experience problems please contact James <james@delx.cjb.net>")
		debug.log("Session: New session created \"%s\" \"%s\" \"%s\" \"%s\"" % (jabberID, username, password, nickname))
	
	
	def removeMe(self):
		""" Safely removes the session object, including sending <presence type="unavailable"/> messages for each legacy related item on the user's contact list """
		# Send offline presence to Jabber ID
		# Delete all objects cleanly
		# Remove this Session object from the pytrans
		
		debug.log("Session: Removing \"%s\"" % (self.jabberID))
		
		self.sendPresence(to=self.jabberID, fro=config.jid, ptype="unavailable")
		self.legacycon.removeMe()
		self.legacycon = None
		
		# Clean up objects
		del self.pytrans.sessions[self.jabberID]
		jabw.JabberConnection.removeMe(self)
		self.pytrans = None
		
		debug.log("Session: Completed removal \"%s\"" % (self.jabberID))
	
	def buildFriendly(self, status):
		""" Constructs a friendly name from the user's registered nick, and their status message """
		if(self.nickname != None and len(self.nickname) > 0):
			friendly = self.nickname
		else:
			friendly = self.jabberID[:self.jabberID.find('@')]
		if(status and len(status) > 0):
			friendly += " - "
			friendly += status
		if(len(friendly) > 127):
			friendly = friendly[:124] + "..."
		debug.log("Session: buildFriendly(%s) returning \"%s\"" % (self.jabberID, friendly))
		return friendly
	
	def messageReceived(self, source, dest, resource, mtype, body):
		""" Sends the message to the legacy translator """
		debug.log("Session: messageReceived(), passing onto legacycon.sendMessage()")
		self.legacycon.sendMessage(dest, body)
	
	def presenceReceived(self, source, resource, priority, ptype, show, status):
		""" Checks resources and priorities so that the highest priority resource always appears as the
		legacy services status. If there are no more resources then the session is deleted """
		r = SessionResource(resource, show, status, priority)
		if(not (ptype in ["unavailable", "error"])):
			# Check to see if this is a new resource, if so, send out the contact list
			debug.log("Session: Presence change received \"%s\"" % (self.jabberID))
			if(self.resourceList.count(r) == 0):
				debug.log("Session: New resource online \"%s\" \"%s\" \"%s\"" % (self.jabberID, resource, priority))
				self.legacycon.newResourceOnline(resource)
			
			# We must check the resource and priority only interested in the resource and priority
			higher = False
			oldList = utils.copyList(self.resourceList)
			self.resourceList = []
			for checkR in oldList:
				if(checkR > r):
					higher = True
				if(checkR == r):
					self.resourceList.append(r)
				else:
					self.resourceList.append(checkR)
				
			if(not higher):
				# No higher resource was found. So we must be the highest!
				# This means we must update the presence on the legacy network
				debug.log("Session: \"%s\" Updating status on legacy network \"%s\" \"%s\"" % (self.jabberID, show, status))
				self.legacycon.setStatus(show, self.buildFriendly(status))
		
		else:
			# This resource has gone offline, so we need to remove it from the list, and find the next highest
			# resource to use to update our status on the legacy network
			debug.log("Session: Unavailable presence received from \"%s\". Looking for other resources" % (self.jabberID))
			lst = utils.copyList(self.resourceList)
			self.resourceList = []
			for i in lst:
				if(r != i):
					self.resourceList.append(i)
			
			highest = None
			for checkR in self.resourceList:
				if(checkR > highest): # comparisons against None are always True
					highest = checkR
			if(highest == None):
				# There are no resources left, so offline this session
				debug.log("Session: No resources left \"%s\"" % (source))
				self.removeMe()
			else:
				# Update the legacy network's status with the highest leftover resource
				debug.log("Session: New resource \"%s\" found for session \"%s\"", (highest, self.jabberID))
				self.legacycon.setStatus(show, self.buildFriendly(status))
	
	
	def subscriptionReceived(self, to, subtype):
		""" Sends the subscription request to the legacy services handler """
		debug.log("Session: \"%s\" subscriptionReceived(), passing onto legacycon.jabberSubscriptionReceived()" % (self.jabberID))
		self.legacycon.jabberSubscriptionReceived(to, subtype)
	








class SessionResource:
	""" A convienence class to allow comparisons of Jabber resources """
	def __init__(self, resource, show=None, status=None, priority=None):
		self.resource = resource
		self.show = show
		self.status = status
		if(priority.__class__ != int):
			self.priority = 0
		else:
			self.priority = priority
	
	def __eq__(self, other):
		return (self.resource == other.resource)
	
	def __cmp__(self, other):
		return self.priority.__cmp__(other.priority)
